/************************************************************************/
/*	Copyright 1987-1991 by Chuck Musciano and Harris Corporation 	*/
/*									*/
/*	Full ownership of this software, and all rights pertaining to 	*/
/*	the for-profit distribution of this software, are retained by 	*/
/*	Chuck Musciano and Harris Corporation.  You are permitted to 	*/
/*	use this software without fee.  This software is provided "as 	*/
/*	is" without express or implied warranty.  You may redistribute 	*/
/*	this software, provided that this copyright notice is retained,	*/
/*	and that the software is not distributed for profit.  If you 	*/
/*	wish to use this software in a profit-making venture, you must 	*/
/*	first license this code and its underlying technology from 	*/
/*	Harris Corporation. 						*/
/*									*/
/*	Bottom line: you can have this software, you can use it, you 	*/
/*	can give it away.  You just can't sell any or all parts of it 	*/
/*	without prior permission from Harris Corporation. 		*/
/************************************************************************/

/************************************************************************/
/*									*/
/*	casio_slave.c		handle low-level BOSS i/o		*/
/*									*/
/************************************************************************/

#include	<stdio.h>
#include	<sys/types.h>
#include	<sys/time.h>
#include	<sys/termios.h>
#include	<sys/file.h>

#include	"manifest.h"
#include	"casio.h"
#include	"protocol.h"

PUBLIC	int	errno;
PUBLIC	char	*sys_errlist[];

PRIVATE	int	baud_rate[] = {B1200, B2400, B4800, B9600};
PRIVATE	int	parity[] = {PARENB, PARENB | PARODD, 0};

PRIVATE	int	casio = -1;
PRIVATE	int	previous_byte = -1;

PRIVATE	FILE	*commands_in;
PRIVATE	FILE	*results_out;

/************************************************************************/
PRIVATE	return_result(a, b, c, d, e, f)

int	a, b, c, d, e, f;

{	char	buf[512];

	if (a == NULL)
	   putc(0, results_out);
	else {
	   sprintf(buf, a, b, c, d, e, f);
	   putc(strlen(buf), results_out);
	   fprintf(results_out, buf);
	   }
	fflush(results_out);
}

/************************************************************************/
PRIVATE	open_casio(port, baud, par)

char	*port;
int	baud;
int	par;

{	struct	termios tbuf;

	if ((casio = open(port, O_RDWR)) == -1) {
	   return_result("could not access %s", port);
	   return;
	   }
	if (ioctl(casio, TCGETS, &tbuf) == -1) {
	   close(casio);
	   casio = -1;
	   return_result("error getting parameters for %s (%s)", port, sys_errlist[errno]);
	   return;
	   }
	tbuf.c_iflag = IGNBRK | IXON | IXOFF;
	tbuf.c_oflag = 0;
	tbuf.c_cflag = baud_rate[baud] | parity[par] | CS8 | CREAD;
	tbuf.c_lflag = 0;
	tbuf.c_cc[VMIN] = 0;
	tbuf.c_cc[VTIME] = 0;
	if (ioctl(casio, TCSETS, &tbuf) == -1) {
	   close(casio);
	   casio = -1;
	   return_result("error setting parameters for %s (%s)", port, sys_errlist[errno]);
	   return;
	   }
	return_result(NULL);
}

/************************************************************************/
PRIVATE	close_casio()

{
	close(casio);
	casio = -1;
	return_result(NULL);
}

/************************************************************************/
PRIVATE	int	read_casio(buf, len, max_time)

byte	*buf;
int	len;

{	int	result, amount;
	fd_set	fd;
	struct	timeval	timeout;

	amount = 0;
	if (previous_byte != -1) {
	   *buf = previous_byte;
	   amount++;
	   previous_byte = -1;
	   }
	timeout.tv_sec = max_time / 1000;
	timeout.tv_usec = (max_time % 1000) * 1000;
	while (amount < len) {
	   FD_ZERO(&fd);
	   FD_SET(casio, &fd);
	   switch (select(getdtablesize(), &fd, NULL, NULL, &timeout)) {
	      case 0  : return_result(NULL);
	      		return(amount);
	      case 1  : result = read(casio, buf + amount, len - amount);
	   	        if (result == -1) {
	   		   return_result("error reading from BOSS: %s", sys_errlist[errno]);
	   		   return(-1);
	   		   }
	   	        else
	   	           amount += result;
	   	        break;
	      default : return_result("error awaiting data from BOSS: %s", sys_errlist[errno]);
	   	        return(-1);
	      }
	   }
	return_result(NULL);
	return(len);
}

/************************************************************************/
PRIVATE	int	write_casio(buf, len)

byte	*buf;
int	len;

{	int	result;

	if ((result = write(casio, buf, len)) != len)
	   return_result("error writing to BOSS: %s", sys_errlist[errno]);
	else
	   return_result(NULL);
	return(result);
}

/************************************************************************/
PRIVATE	enable_flow_control()

{	struct	termios tbuf;

	if (casio != -1)
	   if (ioctl(casio, TCGETS, &tbuf) == -1)
	      return_result("error getting parameters for BOSS in enable_flow_control: %s", sys_errlist[errno]);
	   else {
	      tbuf.c_iflag |= IXON | IXOFF;
	      if (ioctl(casio, TCSETS, &tbuf) == -1)
	         return_result("error setting parameters for BOSS in enable_flow_control: %s", sys_errlist[errno]);
	      else
	         return_result(NULL);
	      }
	else
	   return_result("attempt to enable flow control without opening BOSS");
}

/************************************************************************/
PRIVATE	disable_flow_control()

{	struct	termios tbuf;

	if (casio != -1)
	   if (ioctl(casio, TCGETS, &tbuf) == -1)
	      return_result("error getting parameters for BOSS in disable_flow_control: %s", sys_errlist[errno]);
	   else {
	      tbuf.c_iflag &= ~(IXON | IXOFF);
	      if (ioctl(casio, TCSETS, &tbuf) == -1)
	         return_result("error setting parameters for BOSS in disable_flow_control: %s", sys_errlist[errno]);
	      else
	         return_result(NULL);
	      }
	else
	   return_result("attempt to disable flow control without opening BOSS");
}

/************************************************************************/
PRIVATE	flush_casio()

{
	if (ioctl(casio, TCFLSH, 2) == -1)
	   return_result("could not flush BOSS output queue: %s", sys_errlist[errno]);
	else
	   return_result(NULL);
}

/************************************************************************/
PRIVATE	unget_byte(b)

byte	b;

{
	if (previous_byte == -1) {
	   previous_byte = b;
	   return_result(NULL);
	   }
	else
	   return_result("unget_byte overrun!");
}

/************************************************************************/
EXPORT	start_slave_handler(commands, results)

int	commands;
int	results;

{	int	buf_size, cmd, baud, parity, len, b, timeout;
	char	*buf;

	if ((commands_in = fdopen(commands, "r")) == NULL) {
	   fprintf(stderr, "Could not fdopen commands: %s", sys_errlist[errno]);
	   exit(1);
	   }
	if ((results_out = fdopen(results, "w")) == NULL) {
	   fprintf(stderr, "Could not fdopen results: %s", sys_errlist[errno]);
	   exit(1);
	   }
	buf = (char *) malloc(buf_size = 1024);
	while ((cmd = getc(commands_in)) != EOF)
	   switch (cmd) {
	      case OPEN_CASIO           : if ((baud = getc(commands_in)) != EOF)
	      				     if ((parity = getc(commands_in)) != EOF)
	      				        if ((len = getc(commands_in)) != EOF)
	      				           if (fread(buf, 1, len, commands_in) == len) {
	      				              if (buf_size < len)
	      				                 buf = (char *) realloc(buf, buf_size = len);
	      				              buf[len] = '\0';
	      				              open_casio(buf, baud, parity);
	      				              continue;
	      				              }
	      				  return_result("Error reading OPEN_CASIO parameters in Casio slave");
	      				  break;
	      case CLOSE_CASIO          : close_casio();
	      				  break;
	      case READ_CASIO           : if (fread(&len, 1, sizeof(int), commands_in) == sizeof(int))
	      				     if (fread(&timeout, 1, sizeof(int), commands_in) == sizeof(int)) {
	      				        if (buf_size < len)
	      				           buf = (char *) realloc(buf, buf_size = len);
	      				        if ((b = read_casio(buf, len, timeout)) != -1) {
	      				           fwrite(&b, 1, sizeof(int), results_out);
	      				           if (b > 0)
	      				              fwrite(buf, 1, b, results_out);
	      				           fflush(results_out);
	      				           continue;
	      				           }
	      				        }
	      				  return_result("Error reading READ_CASIO parameters in Casio slave");
	      				  break;
	      case WRITE_CASIO          : if (fread(&len, 1, sizeof(int), commands_in) == sizeof(int)) {
	      				     if (buf_size < len)
	      				        buf = (char *) realloc(buf, buf_size = len);
	      				     if (fread(buf, 1, len, commands_in) == len) {
	      				        if ((b = write_casio(buf, len)) != -1) {
	      				           fwrite(&b, 1, sizeof(int), results_out);
	      				           fflush(results_out);
	      				           continue;
	      				           }
	      				        }
	      				     }
	      				  return_result("Error reading WRITE_CASIO parameters in Casio slave");
	      				  break;
	      case ENABLE_FLOW_CONTROL  : enable_flow_control();
	      				  break;
	      case DISABLE_FLOW_CONTROL : disable_flow_control();
	      				  break;
	      case FLUSH_CASIO          : flush_casio();
	      				  break;
	      case UNGET_BYTE           : if ((b = getc(commands_in)) != EOF)
	      				     unget_byte(b);
	      				  break;
	      case WAIT_CASIO           : if (fread(&timeout, 1, sizeof(int), commands_in) == sizeof(int)) {
	      				     usleep(timeout * 1000);
	      				     return_result(NULL);
	      				     }
	      				  else
	      				     return_result("Error reading WAIT_CASIO parameters in Casio slave");
	      				  break;
	      default                   : return_result("Unknown protocol command: %d", cmd);
	      				  break;
	      }
	close(commands);
	close(results);
}
